#!/usr/bin/env python

import hydra
from omegaconf import DictConfig
from collections import defaultdict

from tqdm import tqdm

from pedl.utils import Entity, get_hgnc_symbol_to_gene_id
from pedl.data_getter import DataGetterPubtator, DataGetterAPI
from pedl.predict import maybe_mapped_entities


@hydra.main(config_path="configs", config_name="build_training_set.yaml", version_base=None)
def build_training_set(cfg: DictConfig):
    """
`build_training_set` is a function that generates a training set from a dataset of triples containing entity pairs and their relationships. The generated training set consists of sentences from scientific articles mentioning these entity pairs.

**Arguments:**

- `cfg` (`DictConfig`): A dictionary containing the configuration settings for this function. The settings include:
  - `triples`: The file path of the dataset containing triples (entity pairs and their relationships).
  - `elastic`: The ElasticSearch settings.
  - `entities.entity_marker`: The entity marker settings.
  - `out`: The output file path.
  - `out_blinded`: The output file path for blinded sentences (entities replaced by placeholders).

**Outputs:**

This function generates two TSV files:

1. `<cfg.out>.tsv` or `<cfg.out>.<cfg.worker_id>`: A TSV file containing sentences with the following columns:
   - Head entity type
   - Head entity CUID
   - Tail entity type
   - Tail entity CUID
   - Relationships
   - Sentence text
   - PubMed ID

2. `<cfg.out_blinded>.tsv` or `<cfg.out_blinded>.<cfg.worker_id>`: A TSV file containing blinded sentences (entities replaced by placeholders) with the same columns as the first output file.
    """
    pair_to_relations = defaultdict(set)

    gene_universe = set()
    chemical_universe = set()
    gene_mapper = get_hgnc_symbol_to_gene_id()

    with open(cfg.triples) as f:
        for line in f:
            fields = line.strip().split("\t")
            if fields:
                type_head, cuid_head, type_tail, cuid_tail, rel = fields
                head = Entity(cuid=cuid_head, type=type_head)
                tail = Entity(cuid=cuid_tail, type=type_tail)

                pair_to_relations[(head, tail)].add(rel)

                if head.type == "Chemical":
                    chemical_universe.add(head.cuid)
                elif head.type == "Gene":
                    head.cuid = gene_mapper.get(head.cuid, head.cuid)
                    gene_universe.add(head.cuid)
                elif head.type == "Protein":
                    head.cuid = gene_mapper.get(head.cuid, head.cuid)
                    gene_universe.add(head.cuid)
                    head.type = "Gene"
                else:
                    raise ValueError(head.type)
                if tail.type == "Chemical":
                    chemical_universe.add(tail.cuid)
                elif tail.type == "Protein":
                    tail.cuid = gene_mapper.get(tail.cuid, tail.cuid)
                    gene_universe.add(tail.cuid)
                    tail.type = "Gene"
                elif tail.type == "Gene":
                    tail.cuid = gene_mapper.get(tail.cuid, tail.cuid)
                    gene_universe.add(tail.cuid)
                else:
                    raise ValueError(tail.type)
    data_getter = DataGetterPubtator(elasticsearch=cfg.elastic,
                                     entity_marker=cfg.entities.entity_marker,
                                     max_size=cfg.max_size,)
    with open(str(cfg.out) + ".tsv", "w") as f, open(str(cfg.out_blinded) + ".tsv", "w") as f_blinded:
        for pair, relations in tqdm(list(pair_to_relations.items()),
                                    desc="Crawling"):
            head, tail = pair[:2]
            sentences = data_getter.get_sentences(head, tail)
            for sentence in sentences:
                f.write("\t".join([head.type, head.cuid, tail.type, tail.cuid, ",".join(relations), sentence.text,
                                   sentence.pmid]) + "\n")
                f_blinded.write("\t".join(
                    [head.type, head.cuid, tail.type, tail.cuid, ",".join(relations), sentence.text_blinded,
                     sentence.pmid]) + "\n")



if __name__ == '__main__':
    build_training_set()



import re
import sys
from collections import defaultdict
from operator import itemgetter
from pathlib import Path
from typing import List
import multiprocessing as mp

import bioc
import numpy as np
from elasticsearch import Elasticsearch, helpers
from tqdm import tqdm

from pedl.utils import SegtokSentenceSplitter, get_pmid, chunks, cache_root, replace_consistently_dict
from pedl.data_getter import DataGetterAPI

INDEX_NAME = "pubtator_masked"


def _add_masked_entities(elastic_doc, mask_types=None) -> None:
    elastic_doc["entities_masked"] = {}
    entities = list(elastic_doc["entities"])

    span_to_replacement = {}

    idx_mask = 1
    for entity in entities:
        spans = elastic_doc["entities"][entity]

        entity_type = entity.split("|")[1]
        replacement = None
        if entity_type in mask_types:
            for span in spans:
                span = tuple(span)
                if span in span_to_replacement:
                    continue
                else:
                    if replacement is None:
                        replacement = "<" + mask_types[entity_type] + str(idx_mask) + "/>"
                        idx_mask += 1
                    # fix spans crossing sentence boundaries
                    if span[1] > len(elastic_doc["text"]):
                        span = (span[0], len(elastic_doc["text"]))

                    span_to_replacement[span] = replacement

    text = elastic_doc["text"]
    text_masked, old_span_to_new = replace_consistently_dict(text, span_to_replacement)

    elastic_doc["text_masked"] = text_masked

    for entity in entities:
        spans = elastic_doc["entities"][entity]
        new_spans = []
        for span in spans:
            span = tuple(span)
            if span in old_span_to_new:
                new_spans.append(old_span_to_new[span])
            else:
                new_spans.append(span)
        elastic_doc["entities_masked"][entity] = new_spans



    # assert len(elastic_doc["entities_masked"]) == len(elastic_doc["entities"])
    # for entity in elastic_doc["entities_masked"]:
    #     assert len(elastic_doc["entities_masked"][entity]) >= len(elastic_doc["entities"][entity])


def _process_pubtator_files(files: List[Path], q: mp.Queue, mask_types=None, entity_marker: dict = None):
    sentence_splitter = SegtokSentenceSplitter(entity_marker=entity_marker)
    for file in files:
        actions = []
        with file.open() as f:
            collection = bioc.load(f)
            for i, document in enumerate(collection.documents):
                pmid, is_fulltext = get_pmid(document)

                for passage in document.passages:
                    if is_fulltext and (
                            passage.infons["type"] == "ref" or # this is a reference
                            passage.infons.get("section_type", "").lower() in {"title", "abstract"}): # this was already added via the abstract
                        continue

                    entities_passage = []
                    for annotation in passage.annotations:
                        entities_ann = DataGetterAPI.get_entities_from_annotation(annotation=annotation,
                                                                                  homologue_mapping={})
                        for entity in entities_ann:
                            for location in annotation.locations:
                                span = (location.offset-passage.offset,
                                        location.end-passage.offset)
                                entities_passage.append((span, entity))
                    entities_passage = sorted(entities_passage, key=itemgetter(0), reverse=True) # reversed because we pop from end, which gives us O(1)

                    if not entities_passage:
                        continue

                    entity = entities_passage.pop()
                    for sentence in sentence_splitter.split(passage.text):
                        if not entity: # we have exhausted all entities
                            break

                        if entity[0][0] > sentence.end_pos: # this sentence contains no entity => skip
                            continue
                        elastic_doc = {
                            "text": sentence.text,
                            "cellline": [],
                            "chemical": [],
                            "disease": [],
                            "gene": [],
                            "proteinmutation": [],
                            "dnamutation": [],
                            "snpmutation": [],
                            "species": [],
                            "entities": defaultdict(list),
                            "pmid": pmid,
                            "span": [sentence.start_pos + passage.offset, sentence.end_pos + passage.offset + passage.offset + passage.offset]
                        }
                        while entity[0][0] < sentence.end_pos:
                            if entity[0][0] - sentence.start_pos < 0:
                                if len(entities_passage) == 0:
                                    entity = None
                                    break
                                else:
                                    entity = entities_passage.pop()
                                continue

                            if entity[1].type.lower() in elastic_doc:
                                elastic_doc[entity[1].type.lower()].append(entity[1].cuid)
                                span = [entity[0][0] - sentence.start_pos, entity[0][1] - sentence.start_pos]
                                elastic_doc["entities"][entity[1].cuid + "|" + entity[1].type].append(span)
                            if len(entities_passage) == 0:
                                entity = None
                                break
                            else:
                                entity = entities_passage.pop()

                        _add_masked_entities(elastic_doc, mask_types)

                        action = {
                            "_index": INDEX_NAME,
                            "_source": elastic_doc
                        }
                        actions.append(action)

        q.put(actions)


def build_index(pubtator_file, n_processes, elastic_search_server, masked_types=None, entity_marker: dict = None,
                password: str = False, ca_certs: str = False):

    n_processes = n_processes or mp.cpu_count()
    if password and ca_certs:
        client = Elasticsearch(elastic_search_server,
                               timeout=3000,
                               basic_auth=("elastic", password),
                               ca_certs=ca_certs)
    elif elastic_search_server:
        client = Elasticsearch(elastic_search_server,
                               timeout=3000,)
    else:
        client = Elasticsearch(timeout=3000,)

    if client.indices.exists(index=INDEX_NAME):
        client.indices.delete(index=INDEX_NAME)

    client.indices.create(
        index=INDEX_NAME,
        mappings={
            "dynamic": False,
            "properties": {
                "cellline": {"type": "keyword"},
                "chemical": {"type": "keyword"},
                "disease": {"type": "keyword"},
                "gene": {"type": "keyword"},
                "proteinmutation": {"type": "keyword"},
                "dnamutation": {"type": "keyword"},
                "snpmutation": {"type": "keyword"},
                "species": {"type": "keyword"},
                "pmid": {"type": "keyword"},
            }
        }
                          )

    ctx = mp.get_context()
    q = ctx.Queue(maxsize=10)
    files = list(pubtator_file.glob("*BioC.XML"))
    processes = []
    for file_chunk in chunks(files, len(files) // n_processes - 1):
        p = ctx.Process(
            target=_process_pubtator_files, args=(file_chunk, q, masked_types, entity_marker)
        )
        p.start()
        processes.append(p)

    pbar = tqdm(desc="Building pubtator elastic index", total=len(files))
    n_files_processed = 0
    while n_files_processed < len(files):
        elastic_docs = q.get()
        num_attempts = 0
        all_errors = []
        while errors := helpers.bulk(client, elastic_docs)[1]:
            num_attempts += 1
            all_errors += errors
            if num_attempts >= 30:
                print("Too many attempts. Errors:")
                print(all_errors)
                sys.exit(1)
        n_files_processed += 1
        pbar.update()

    for p in processes:
        p.join()




from collections import defaultdict

from torch import nn
import torch
from transformers import BertPreTrainedModel, AutoModel, BertModel, \
    DataCollatorWithPadding
from tqdm import tqdm

from pedl.utils import chunks


class BertForDistantSupervision(BertPreTrainedModel):
    def __init__(self,
                 config,
                 tokenizer,
                 *inputs,
                 use_cls: bool = False,
                 use_starts: bool = False,
                 use_ends: bool = False,
                 entity_marker: dict = None,
                 num_label: int = 7,
                 **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.num_labels = num_label
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.tokenizer = tokenizer
        self.use_cls = use_cls
        self.use_starts = use_starts
        self.use_ends = use_ends
        self.init_weights()
        if entity_marker:
            self.entity_marker = entity_marker
        else:
            self.entity_marker = {"head_start": '<e1>',
                                  "head_end": '</e1>',
                                  "tail_start": '<e2>',
                                  "tail_end": '</e2>'}
        seq_rep_size = 0
        if use_cls:
            seq_rep_size += self.bert.config.hidden_size
        if use_starts:
            seq_rep_size += 2 * self.bert.config.hidden_size
        if use_ends:
            seq_rep_size += 2 * self.bert.config.hidden_size
        self.classifier = nn.Linear(seq_rep_size, self.num_labels)

    def forward(self, input_ids, attention_mask, use_max=False, **kwargs):
        bert_out = self.bert(input_ids, attention_mask=attention_mask)
        seq_emb = bert_out.last_hidden_state
        seq_reps = []
        if self.use_cls:
            seq_reps.append(seq_emb[:, 0])
        if self.use_starts:
            head_start_idx = torch.where(
                input_ids
                == self.tokenizer.convert_tokens_to_ids(self.entity_marker['head_start'])
            )
            tail_start_idx = torch.where(
                input_ids
                == self.tokenizer.convert_tokens_to_ids(self.entity_marker['tail_start'])
            )
            head_start_rep = seq_emb[head_start_idx]
            tail_start_rep = seq_emb[tail_start_idx]
            start_pair_rep = torch.cat([head_start_rep, tail_start_rep], dim=1)
            seq_reps.append(start_pair_rep)

        if self.use_ends:
            head_end_idx = torch.where(
                input_ids
                == self.tokenizer.convert_tokens_to_ids(self.entity_marker['head_end'])
            )
            tail_end_idx = torch.where(
                input_ids
                == self.tokenizer.convert_tokens_to_ids(self.entity_marker['tail_end'])
            )
            head_end_rep = seq_emb[head_end_idx]
            tail_end_rep = seq_emb[tail_end_idx]
            end_pair_rep = torch.cat([head_end_rep, tail_end_rep], dim=1)
            seq_reps.append(end_pair_rep)

        seq_reps = torch.cat(seq_reps, dim=1)
        seq_emb = self.dropout(seq_reps)

        logits = self.classifier(seq_emb)
        if use_max:
            alphas = torch.max(logits, dim=1)[0]
        else:
            alphas = torch.logsumexp(logits, dim=1)
        meta = {
            'alphas': alphas,
            'alphas_by_rel': logits,
        }

        if use_max:
            bag_logits = torch.max(logits, dim=0)[0]
        else:
            bag_logits = torch.logsumexp(logits, dim=0)

        return bag_logits, meta

    def collate_fn(self, batch):
        if "sentences" not in batch[0]:
            return batch[0]
        collated_batch = {}
        collator = DataCollatorWithPadding(tokenizer=self.tokenizer)
        for k, v in batch[0].items():
            if k != "encoding":
                collated_batch[k] = v
        collated_batch["encoding"] = collator(batch[0]["encoding"])

        return collated_batch

    def forward_batched(self, input_ids, attention_mask, batch_size, **kwargs):
        logits_all = []
        meta_all = defaultdict(list)

        if not batch_size:
            return self.forward(input_ids, attention_mask)

        indices = torch.arange(len(input_ids))
        for indices_batch in tqdm(list(chunks(indices, batch_size))):
            input_ids_batch = input_ids[indices_batch]
            attention_mask_batch = attention_mask[indices_batch]
            logits, meta = self.forward(input_ids=input_ids_batch,
                                        attention_mask=attention_mask_batch)
            logits_all.append(logits)
            for k, v in meta.items():
                meta_all[k].append(v)

        logits_all = torch.cat(logits_all)
        for k, v in meta_all.items():
            meta_all[k] = torch.cat(v)

        return logits_all, meta_all




#!/usr/bin/env python

import sys
import hydra
from pathlib import Path
from omegaconf import DictConfig
from pedl import pubtator_elasticsearch


@hydra.main(config_path="configs", config_name="rebuild_pubtator_index.yaml", version_base=None)
def rebuild_pubtator_index(cfg: DictConfig):
    really_continue = input("This will delete the pubtator index and rebuild it. Do you want to continue? Type 'yes':\n").strip()
    if really_continue != "yes":
        print(f"Your input was '{really_continue}', not 'yes'. Aborting.")
        sys.exit(0)
    pubtator_elasticsearch.build_index(pubtator_file=Path(cfg.pubtator_file),
                                       n_processes=cfg.n_processes,
                                       elastic_search_server=cfg.elastic.server,
                                       masked_types=cfg.type.entity_to_mask,
                                       entity_marker=cfg.entities.entity_marker,
                                       password=cfg.elastic.password,
                                       ca_certs=cfg.elastic.ca_certs
                                       )



import gzip
import itertools
import json
import logging
import os
import re
import shutil
import tempfile
import urllib
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from operator import itemgetter
from pathlib import Path
from typing import Union, Optional, List, Set, Tuple, Dict
from urllib.parse import urlparse
from urllib.request import urlopen

import bioc
import numpy as np
import requests
from pytorch_lightning.utilities import rank_zero_only
from segtok.segmenter import split_multi
from tqdm import tqdm as _tqdm, tqdm
from transformers.file_utils import default_cache_path

cache_root = Path(
    os.getenv("PEDL_CACHE", Path(default_cache_path).parent.parent / "pedl")
)
if not cache_root.exists():
    os.makedirs(cache_root, exist_ok=True)

root = Path(__file__).parent

class Sentence:
    def __init__(
        self,
        text: str,
        start_pos: int,
        pmid: Optional[str] = None,
        text_blinded: Optional[str] = None,
        entity_marker: dict = None,
    ):
        self.text = text
        self.pmid = pmid
        self.start_pos = start_pos
        self.end_pos = start_pos + len(text)
        self.text_blinded = text_blinded
        if entity_marker:
            self.em = entity_marker
        else:
            self.em = {"head_start": '<e1>',
                       "head_end": '</e1>',
                       "tail_start": '<e2>',
                       "tail_end": '</e2>'}

    def __str__(self):
        return self.text

    def __repr__(self):
        return str(self)

    def get_unmarked_text(self):
        return (
            self.text.replace(self.em["head_start"], "")
            .replace(self.em["head_end"], "")
            .replace(self.em["tail_start"], "")
            .replace(self.em["tail_end"], "")
        )

    def __hash__(self):
        return hash((self.pmid, self.text))

    def __eq__(self, other):
        return (self.pmid, self.text) == (other.pmid, other.text)


class SegtokSentenceSplitter:
    def __init__(self, entity_marker: dict = None):
        if entity_marker:
            self.entity_marker = entity_marker
        else:
            self.entity_marker = {"head_start": '<e1>',
                                  "head_end": '</e1>',
                                  "tail_start": '<e2>',
                                  "tail_end": '</e2>'}
    """
    For further details see: https://github.com/fnl/segtok
    """

    def split(self, text: str) -> List[Sentence]:
        sentences = []
        offset = 0

        plain_sentences = split_multi(text)
        for sentence in plain_sentences:
            sentence_offset = text.find(sentence, offset)

            if sentence_offset == -1:
                raise AssertionError(
                    f"Can't find offset for sentences {plain_sentences} "
                    f"starting from {offset}"
                )

            sentences += [
                Sentence(
                    text=sentence,
                    start_pos=sentence_offset,
                    entity_marker=self.entity_marker
                )
            ]

            offset += len(sentence)

        return sentences


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def get_pmid(document: bioc.BioCDocument) -> Tuple[str, int]:
    infons = document.passages[0].infons
    if "article-id_pmid" in infons:
        pmid = document.passages[0].infons["article-id_pmid"]
        is_fulltext = int(document.id == infons.get("article-id_pmc", None))
    else:
        pmid = document.id
        is_fulltext = 0

    return pmid, is_fulltext


class Tqdm:
    # These defaults are the same as the argument defaults in tqdm.
    default_mininterval: float = 0.1

    @staticmethod
    def set_default_mininterval(value: float) -> None:
        Tqdm.default_mininterval = value

    @staticmethod
    def set_slower_interval(use_slower_interval: bool) -> None:
        """
        If ``use_slower_interval`` is ``True``, we will dramatically slow down ``tqdm's`` default
        output rate.  ``tqdm's`` default output rate is great for interactively watching progress,
        but it is not great for log files.  You might want to set this if you are primarily going
        to be looking at output through log files, not the terminal.
        """
        if use_slower_interval:
            Tqdm.default_mininterval = 10.0
        else:
            Tqdm.default_mininterval = 0.1

    @staticmethod
    def tqdm(*args, **kwargs):
        new_kwargs = {"mininterval": Tqdm.default_mininterval, **kwargs}

        return _tqdm(*args, **new_kwargs)


def get_logger(name=__name__, level=logging.INFO) -> logging.Logger:
    """Initializes multi-GPU-friendly python logger."""

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in ("debug", "info", "warning", "error", "exception", "fatal", "critical"):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def get_from_cache(url: str, cache_dir: Path = None) -> Path:
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    cache_dir.mkdir(parents=True, exist_ok=True)

    filename = re.sub(r".+/", "", url)
    # get cache path to put the file
    cache_path = cache_dir / filename
    if cache_path.exists():
        return cache_path

    # make HEAD request to check ETag
    response = requests.head(url, headers={"User-Agent": "Flair"}, allow_redirects=True)
    if response.status_code != 200:
        raise IOError(
            f"HEAD request failed for url {url} with status code {response.status_code}."
        )

    # add ETag to filename if it exists
    # etag = response.headers.get("ETag")

    if not cache_path.exists():
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        fd, temp_filename = tempfile.mkstemp()

        # GET file object
        req = requests.get(url, stream=True, headers={"User-Agent": "PEDL"})
        content_length = req.headers.get("Content-Length")
        total = int(content_length) if content_length is not None else None
        progress = Tqdm.tqdm(unit="B", total=total)
        with open(temp_filename, "wb") as temp_file:
            for chunk in req.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    progress.update(len(chunk))
                    temp_file.write(chunk)

        progress.close()

        shutil.copyfile(temp_filename, str(cache_path))
        os.close(fd)
        os.remove(temp_filename)

    return cache_path


def cached_path(url_or_filename: str, cache_dir: Union[str, Path]) -> Path:
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    if type(cache_dir) is str:
        cache_dir = Path(cache_dir)
    dataset_cache = Path(cache_root) / cache_dir

    parsed = urlparse(url_or_filename)

    if parsed.scheme in ("http", "https"):
        # URL, so get it from the cache (downloading if necessary)
        return get_from_cache(url_or_filename, dataset_cache)
    elif parsed.scheme == "" and Path(url_or_filename).exists():
        # File, and it exists.
        return Path(url_or_filename)
    elif parsed.scheme == "":
        # File, but it doesn't exist.
        raise FileNotFoundError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError(
            "unable to parse {} as a URL or as a local path".format(url_or_filename)
        )


def unpack_file(file: Path, unpack_to: Path, mode: str = None, keep: bool = True):
    """
    Unpacks a file to the given location.

    :param file Archive file to unpack
    :param unpack_to Destination where to store the output
    :param mode Type of the archive (zip, tar, gz, targz, rar)
    :param keep Indicates whether to keep the archive after extraction or delete it
    """
    if mode == "zip" or (mode is None and str(file).endswith("zip")):
        from zipfile import ZipFile

        with ZipFile(file, "r") as zipObj:
            # Extract all the contents of zip file in current directory
            zipObj.extractall(unpack_to)

    elif mode == "targz" or (
        mode is None and str(file).endswith("tar.gz") or str(file).endswith("tgz")
    ):
        import tarfile

        with tarfile.open(file, "r:gz") as tarObj:
            tarObj.extractall(unpack_to)

    elif mode == "tar" or (mode is None and str(file).endswith("tar")):
        import tarfile

        with tarfile.open(file, "r") as tarObj:
            tarObj.extractall(unpack_to)

    elif mode == "gz" or (mode is None and str(file).endswith("gz")):
        import gzip

        with gzip.open(str(file), "rb") as f_in:
            with open(str(unpack_to), "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

    elif mode == "rar" or (mode is None and str(file).endswith("rar")):
        import patoolib

        patoolib.extract_archive(str(file), outdir=unpack_to, interactive=False)

    else:
        if mode is None:
            raise AssertionError(f"Can't infer archive type from {file}")
        else:
            raise AssertionError(f"Unsupported mode {mode}")

    if not keep:
        os.remove(str(file))


def replace_consistently(offset, length, replacement, text, offsets):
    delta_len = len(replacement) - length
    new_text = text[:offset] + replacement + text[offset + length :]
    new_offsets = offsets.copy()
    new_offsets[offsets >= offset + length] += delta_len

    return new_text, new_offsets


def insert_consistently(offset, insertion, text, starts, ends, idx_changed):
    new_text = text[:offset] + insertion + text[offset:]
    new_starts = starts.copy()
    new_ends = ends.copy()
    not_idx_changed_mask = np.ones(len(starts), dtype=bool)
    not_idx_changed_mask[idx_changed] = False
    new_starts[(starts >= offset) & not_idx_changed_mask] += len(insertion)
    new_ends[ends > offset] += len(insertion)

    return new_text, new_starts, new_ends


def delete_consistently(from_idx, to_idx , text, starts, ends):
    assert to_idx >= from_idx
    new_text = text[:from_idx] + text[to_idx:]
    new_starts = starts.copy()
    new_ends = ends.copy()
    new_starts[(from_idx <= starts) & (starts <= to_idx)] = from_idx
    new_ends[(from_idx <= ends) & (ends <= to_idx)] = from_idx
    new_starts[starts > to_idx] -= (to_idx - from_idx)
    new_ends[ends > to_idx] -= (to_idx - from_idx)

    return new_text, new_starts, new_ends


def replace_consistently_dict(text, span_to_replacement):
    old_offset_to_new = {}
    current_offset = 0
    new_text = ""

    # Sort spans by starting index
    sorted_spans = sorted(span_to_replacement.items(), key=lambda x: x[0][0])

    for (start, end), substitution in sorted_spans:
        # Add the text before the current span
        new_text += text[current_offset:start]

        # Add the substitution and update the new offsets
        old_offset_to_new[(start, end)] = (len(new_text),  len(substitution) + len(new_text))
        new_text += substitution

        # Update the current offset
        current_offset = end

    # Add the remaining text
    new_text += text[current_offset:]

    # check results
    for (start, end), (new_start, new_end) in old_offset_to_new.items():
        assert new_text[new_start:new_end] == span_to_replacement[(start, end)]

    return new_text, old_offset_to_new




def get_homologue_mapping(
    expand_species_names: List[str], protein_universe: Set[str]
) -> Dict[str, Set[str]]:
    """
    Get mapping from genes in `expand_species` that are not in `protein_universe` to all
    homologous gene_id in `protein_universe`
    """
    species_to_tax_id = {
        "human": "9606",
        "mouse": "10090",
        "rat": "10116",
        "zebrafish": "7955",
    }
    expand_species = [species_to_tax_id[i] for i in expand_species_names]
    homologue_mapping = defaultdict(set)
    gene_id_to_cluster_id = {}
    cluster_id_to_gene_ids = defaultdict(set)

    with open(root / "data" / "HOM_AllOrganism.rpt") as f:
        next(f)
        for line in f:
            fields = line.strip().split("\t")
            if not fields:
                continue
            cluster_id = fields[0]
            taxon_id = fields[2]
            gene_id = fields[4]
            if taxon_id in expand_species or gene_id in protein_universe:
                cluster_id_to_gene_ids[cluster_id].add(gene_id)
            if gene_id in protein_universe:
                gene_id_to_cluster_id[gene_id] = cluster_id

        for gene_id in protein_universe:
            if gene_id in gene_id_to_cluster_id:
                cluster_id = gene_id_to_cluster_id[gene_id]
                for mapped_gene_id in cluster_id_to_gene_ids[cluster_id]:
                    if mapped_gene_id not in protein_universe:
                        homologue_mapping[mapped_gene_id].add(gene_id)

    return dict(homologue_mapping)


@dataclass(unsafe_hash=True)
class Entity:
    cuid: str
    type: str

    def to_json(self):
        return [self.cuid, self.type]

    def __str__(self):
        return f"{self.cuid}|{self.type}"

    def __lt__(self, other):
        if not isinstance(other, Entity):
            raise TypeError

        return str(self) < str(other)

@dataclass
class DatasetMetaInformation:
    label_to_id: dict
    id_to_label: dict
    pair_types: set
    name: str
    type: str

    def to(self, device):
        return self


def get_geneid_to_name():
    with open(root / "data" / "geneid_to_name.json") as f:
        return json.load(f)


def get_gene_mapping(from_db: str, to_db: str):
    final_mapping = defaultdict(set)
    uniprot_to_from_db = defaultdict(set)
    uniprot_to_to_db = defaultdict(set)
    uniprot_mapping_file = cached_path(
        "https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/by_organism/HUMAN_9606_idmapping.dat.gz",
        cache_dir=cache_root,
    )
    with gzip.open(uniprot_mapping_file) as f:
        for line in f:
            uniprot_id, other_db, other_db_id = line.decode().strip().split("\t")
            if other_db == from_db:
                uniprot_to_from_db[uniprot_id].add(other_db_id)
            if other_db == to_db:
                uniprot_to_to_db[uniprot_id].add(other_db_id)

        for k in uniprot_to_from_db:
            from_ids = uniprot_to_from_db[k]
            to_ids = uniprot_to_to_db[k]

            for from_id in from_ids:
                for to_id in to_ids:
                    final_mapping[from_id].add(to_id)

    return dict(final_mapping)



def overlaps(a, b):
    a = [int(i) for i in a]
    b = [int(i) for i in b]
    return max(0, min(a[1], b[1]) - max(a[0], b[0]))


def get_hgnc_symbol_to_gene_id():
    hgnc_symbol_to_gene_id = {}
    url = "http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/tsv/hgnc_complete_set.txt"
    with open(cached_path(url, cache_dir=cache_root), encoding="utf8") as f:
        next(f)
        for line in f:
            fields = line.strip().split("\t")
            if len(fields) > 18:
                symbol = fields[1]
                gene_id = fields[18]
                hgnc_symbol_to_gene_id[symbol] = gene_id

    return hgnc_symbol_to_gene_id


def get_mesh_id_to_chem_name():
    resp = cached_path("http://ctdbase.org/reports/CTD_chemicals.tsv.gz", cache_dir=cache_root)
    mesh_id_to_chem_name = {}
    with gzip.open(resp) as f:
        lines = f.readlines()
        for line in lines:
            fields = line.decode('utf-8').strip().split("\t")
            if fields[0].startswith('# ChemicalName'):
                continue
            if len(fields) > 5:
                symbol = fields[0]
                mesh_id = fields[1]
                mesh_id_to_chem_name[symbol] = mesh_id
    return mesh_id_to_chem_name


def maybe_mapped_entities(entities, normalized_entity_ids, use_ids=False):
    maybe_mapped_entities = []
    if use_ids:
        for entity in entities:
            if re.match(r"^[A-Z]\d+", entity):
                entity = 'MESH:' + entity
            else:
                if not entity.startswith('MESH:'):
                    entity = entity[0:4].upper() + entity[4:]
            maybe_mapped_entities.append(entity)
    else:
        for entity in entities:
            maybe_mapped_entities.append(normalized_entity_ids[entity])
    return maybe_mapped_entities


def doc_synced(final_path, type):
    local_time = datetime.utcfromtimestamp(os.path.getmtime(final_path)).strftime('%Y-%m-%d %H:%M')
    weburl = urllib.request.urlopen("https://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral")
    url_lines = weburl.readlines()
    if type == 'gene':
        # get the Last modified date of gene2pubtatorcentral.gz
        x = str(url_lines[23]).split('</a>')[-1]
        x = " ".join(list(filter(None, x.split(' ')))[0:2])
    else:
        # get the Last modified date of chemical2pubtatorcentral.gz
        x = str(url_lines[19]).split('</a>')[-1]
        x = " ".join(list(filter(None, x.split(' ')))[0:2])
    if x > local_time:
        really_continue = input(f"Your local {type}2pubtatorcentral is outdated. Do you want to download the lastes version? Type 'yes':\n")
        if really_continue == "yes":
            print("Downloading gene2pubtatorcentral...")
            path = cached_path(
                "https://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/gene2pubtatorcentral.gz",
                "data",
            )
            unpack_file(path, final_path)



import logging
import random
from pathlib import Path
from typing import List, Tuple, Set, Optional, Dict

import hydra.utils
import torch
from torch.utils.data import Dataset, Sampler
from transformers import AutoTokenizer
from segtok.segmenter import split_single

from pedl.utils import Entity
from pedl.data_getter import DataGetterAPI

logger = logging.getLogger(__name__)


class PEDLDataset(Dataset):
    def __init__(
        self,
        heads: List[Entity],
        tails: List[Entity],
        skip_pairs: Set[Tuple[str, str]],
        base_model: str,
        data_getter: DataGetterAPI,
        max_length: int,
        pair_side_information: str = None,
        entity_side_information: str = None,
        relations: Optional[List[Set[str]]] = None,
        max_bag_size: Optional[int] = None,
        mask_entities: bool = True,
        sentence_max_length: Optional[int] = None,
        entity_marker: dict = None,
        entity_to_mask: dict = None,
        label_to_id: dict = None,
        use_starts: bool = False,
        use_ends: bool = False,
    ):
        self.label_to_id = label_to_id
        self.id_to_label = {v: k for k, v in label_to_id.items()}
        self.heads = heads
        self.tails = tails
        self.max_bag_size = max_bag_size
        self.tokenizer = AutoTokenizer.from_pretrained(str(base_model))
        if entity_marker:
            self.entity_marker = entity_marker
        else:
            self.entity_marker = {"head_start": '<e1>',
                                  "head_end": '</e1>',
                                  "tail_start": '<e2>',
                                  "tail_end": '</e2>'}

        self.tokenizer.add_special_tokens(
            {
                "additional_special_tokens": list(entity_marker.values())
            })
        # if entity_to_mask:
            # self.tokenizer.add_special_tokens(
            #     {
            #         "additional_special_tokens": [f"<{entity_to_mask['Gene']}{i}/>" for i in range(1, 47)]
            #     }
            # )

        self.n_classes = len(self.label_to_id)
        self.data_getter = data_getter
        self.relations = relations
        self.mask_entities = mask_entities
        self.sentence_max_length = sentence_max_length
        self.skip_pairs = skip_pairs
        self.max_length = max_length
        self.pair_to_side_information = {}
        if pair_side_information:
            self.pair_to_side_information = self.get_side_information(pair_side_information)
        self.entity_to_side_information = {}
        if entity_side_information:
            self.entity_to_side_information = self.get_side_information(entity_side_information)


    def __len__(self):
        return len(self.heads) * len(self.tails)

    def __getitem__(self, idx) -> Optional[Dict]:
        if torch.is_tensor(idx):
            idx = idx.item()

        head = self.heads[idx // len(self.tails)]
        tail = self.tails[idx % len(self.tails)]

        if (str(head), str(tail)) in self.skip_pairs:
            return {"pair": (head, tail)}

        sentences = self.data_getter.get_sentences(head, tail)

        if self.sentence_max_length:
            sentences = [s for s in sentences if len(s.text) < self.sentence_max_length]

        # deduplicate sentences
        sentences = list(set(sentences))

        if not sentences:
            return {"pair": (head, tail)}

        if self.relations:
            labels = torch.zeros(len(self.label_to_id))
            for relation in self.relations[idx]:
                labels[self.label_to_id[relation]] = 1
        else:
            labels = None

        if self.max_bag_size and len(sentences) > self.max_bag_size:
            sentences = random.sample(sentences, self.max_bag_size)

        if self.mask_entities:
            texts = [s.text_blinded for s in sentences]
        else:
            texts = [s.text for s in sentences]

        encoding = self.tokenizer.batch_encode_plus(texts, max_length=self.max_length,
                                                    truncation=True)
        filtered_encoding = {k: [] for k in encoding}
        filtered_sentences = []
        # move texts with missing entity markers
        for i, text in enumerate(texts):
            if (self.entity_marker["head_start"] in text and self.entity_marker["head_end"] in text and
                    self.entity_marker["tail_start"] in text and self.entity_marker["tail_end"] in text):
                for k in encoding:
                    filtered_encoding[k].append(encoding[k][i])
                filtered_sentences.append(sentences[i])
            else:
                logger.warning(f"Missing entity markers in sentence: {sentences[i]}")

        sample = {
            "encoding": filtered_encoding,
            "labels": labels,
            "is_direct": False,
            "sentences": filtered_sentences,
            "pair": (head, tail)
        }

        return sample

    def get_id(self, label):
        return self.label_to_id[label]

    def get_label(self, label_id):
        return self.id_to_label[label_id]

    @staticmethod
    def get_side_information(file_name):
        side_information = {}
        with open(hydra.utils.to_absolute_path(Path("data") / "side_information" / file_name)) as f:
            for line in f:
                cuid, side_info = line.strip("\n").split("\t")
                side_information[cuid] = side_info
        return side_information



import logging
import os
import sys
import uuid
import hydra
from omegaconf import DictConfig
from pathlib import Path
from typing import Tuple, Set

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from pedl.data_getter import DataGetterPubtator, DataGetterAPI
from pedl.datasets.pedl_dataset import PEDLDataset
from pedl.bert_for_distant_supervision import BertForDistantSupervision
from pedl.utils import get_hgnc_symbol_to_gene_id, get_geneid_to_name, Entity, maybe_mapped_entities, \
    get_mesh_id_to_chem_name


PREFIX_PROCESSED_PAIRS = ".pairs_processed"

ID_MAPPING = {'Chemical': get_mesh_id_to_chem_name,
              'Gene': get_hgnc_symbol_to_gene_id}


def get_processed_pairs(dir_out: Path) -> Set[Tuple[str, str]]:
    processed_pairs = set()
    for file in dir_out.glob(PREFIX_PROCESSED_PAIRS + "*"):
        with file.open() as f:
            for line in f.read().split("\n"):
                processed_pairs.add(tuple(line.split("\t")))
    return processed_pairs


@torch.no_grad()
@hydra.main(config_path="configs", config_name="predict.yaml", version_base=None)
def predict(cfg: DictConfig):

    head_id_to_entity = ID_MAPPING[cfg.type.head_type]()
    tail_id_to_entity = ID_MAPPING[cfg.type.tail_type]()

    if cfg.verbose:
        logging.basicConfig(level=logging.INFO)

    e1_is_all = cfg.e1 == "all"
    e2_is_all = cfg.e2 == "all"

    e1s = get_entity_list(cfg.e1, head_id_to_entity)
    e2s = get_entity_list(cfg.e2, tail_id_to_entity)

    if e1_is_all:
        maybe_mapped_e1s = maybe_mapped_entities(e1s, head_id_to_entity, False)
    else:
        maybe_mapped_e1s = maybe_mapped_entities(e1s, head_id_to_entity, cfg.use_ids)

    if e2_is_all:
        maybe_mapped_e2s = maybe_mapped_entities(e2s, tail_id_to_entity, False)
    else:
        maybe_mapped_e2s = maybe_mapped_entities(e2s, tail_id_to_entity, cfg.use_ids)



    heads = [Entity(cuid, cfg.type.head_type) for cuid in maybe_mapped_e1s]
    tails = [Entity(cuid, cfg.type.tail_type) for cuid in maybe_mapped_e2s]

    if cfg.type.num_workers > 1:
        heads = sorted(heads)
        heads = [head for i, head in enumerate(heads) if i % cfg.num_workers == cfg.type.worker_id]

    processed_pairs = get_processed_pairs(Path(cfg.out))

    geneid_to_name = get_geneid_to_name()
    if len(heads) * len(tails) > 100 and not cfg.elastic.server:
        print(f"Using PEDL without a local PubTator copy is only supported for small queries up to 100 protein pairs. Your query contains {len(heads) * len(tails)} pairs. Aborting.")
        sys.exit(1)

    if not cfg.device:
        if torch.cuda.is_available():
            cfg.device = "cuda"
        else:
            cfg.device = "cpu"

    if cfg.elastic.server:
        data_getter = DataGetterPubtator(elasticsearch=cfg.elastic,
                                         entity_marker=cfg.entities.entity_marker,
                                         entity_to_mask=cfg.type.entity_to_mask,
                                         )
    else:
        if cfg.type.head_type == cfg.type.tail_type:
            gene_universe = set(maybe_mapped_e1s + maybe_mapped_e2s)
            chem_universe = None
        else:
            gene_universe = set(maybe_mapped_e2s)
            chem_universe = set(maybe_mapped_e1s)
        data_getter = DataGetterAPI(gene_universe=gene_universe,
                                    chemical_universe=chem_universe,
                                    expand_species=cfg.type.expand_species,
                                    entity_to_mask=cfg.type.entity_to_mask,
                                    entity_marker=cfg.entities.entity_marker
                                    )
    if cfg.skip_processed:
        skip_pairs = processed_pairs
    else:
        skip_pairs = {}

    dataset = PEDLDataset(heads=heads,
                          tails=tails,
                          skip_pairs=skip_pairs,
                          base_model=cfg.type.model_name,
                          data_getter=data_getter,
                          sentence_max_length=500,
                          max_bag_size=cfg.max_bag_size,
                          entity_marker=cfg.entities.entity_marker,
                          max_length=cfg.type.max_sequence_length,
                          label_to_id=cfg.type.label_to_id,
                          entity_to_mask=cfg.type.entity_to_mask,
                          )
    model = BertForDistantSupervision.from_pretrained(cfg.type.model_name,
                                                      tokenizer=dataset.tokenizer,
                                                      local_model=cfg.type.local_model,
                                                      use_cls=cfg.type.use_cls,
                                                      use_starts=cfg.type.use_starts,
                                                      use_ends=cfg.type.use_ends,
                                                      num_label=cfg.type.num_labels)
    if "cuda" in cfg.device:
        model.bert = nn.DataParallel(model.bert)
    model.eval()
    model.to(cfg.device)
    model.config.e1_id = dataset.tokenizer.convert_tokens_to_ids(cfg.entities.entity_marker["head_start"])
    model.config.e2_id = dataset.tokenizer.convert_tokens_to_ids(cfg.entities.entity_marker["tail_start"])

    os.makedirs(cfg.out, exist_ok=True)

    dataloader = DataLoader(dataset, num_workers=cfg.type.num_workers, batch_size=cfg.type.batch_size,
                            collate_fn=model.collate_fn, prefetch_factor=100)
    with (Path(cfg.out) / f"{PREFIX_PROCESSED_PAIRS}_{uuid.uuid4()}").open("w") as f_pairs_processed:
        for datapoint in tqdm(dataloader, desc="Reading"):
            head, tail = datapoint["pair"]
            if "sentences" not in datapoint:
                f_pairs_processed.write(f"{head}\t{tail}\n")
                continue

            name1 = head.cuid.replace(":", "_")
            name2 = tail.cuid.replace(":", "_")

            file_out = Path(cfg.out) / f"{name1}-_-{name2}.txt"

            if head == tail:
                f_pairs_processed.write(f"{head}\t{tail}\n")
                continue

            if "cuda" in cfg.device:
                with torch.cuda.amp.autocast():
                    x, meta = model.forward_batched(**datapoint["encoding"],
                                                    batch_size=cfg.type.batch_size)
                    probs = torch.sigmoid(meta["alphas_by_rel"])
            else:
                x, meta = model.forward_batched(**datapoint["encoding"],
                                                batch_size=cfg.type.batch_size)
                probs = torch.sigmoid(meta["alphas_by_rel"])

            if (probs < cfg.cutoff).all():
                f_pairs_processed.write(f"{head}\t{tail}\n")
                continue

            with file_out.open("w") as f:
                for max_score in torch.sort(probs.view(-1), descending=True)[0]:
                    if max_score.item() < cfg.cutoff:
                        continue
                    for i, j in zip(*torch.where(probs == max_score)):
                        label = dataset.get_label(j.item())
                        sentence = datapoint["sentences"][i]
                        f.write(f"{label}\t{max_score.item():.2f}\t{sentence.pmid}\t{sentence.text}\tPEDL\n\n")
            f_pairs_processed.write(f"{head}\t{tail}\n")


def get_entity_list(entity, normalized_entity_ids):
    if isinstance(entity, int):
        entity = str(entity)
    if isinstance(entity, str):
        entity = entity.split()
    if len(entity) == 1 and os.path.exists(entity[0]):
        with open(entity[0]) as f:
            p1s = f.read().strip().split("\n")
    elif len(entity) == 1 and entity[0] == "all":
        p1s = sorted(normalized_entity_ids.keys())
    else:
        p1s = [str(e) for e in entity]
    return p1s


if __name__ == "__main__":
    predict()


import gzip
import json
from collections import defaultdict
from typing import Dict, List, Tuple

import indra
from indra import statements
from indra.sources.biopax.api import process_owl_str
from tqdm import tqdm

from pedl.utils import cache_root, cached_path, root, get_gene_mapping


class PathwayCommonsDB:
    def __init__(self, name, gene_universe):
        url = f"https://www.pathwaycommons.org/archives/PC2/v12/PathwayCommons12.{name}.BIOPAX.owl.gz"
        self.name = name
        self.gene_universe = set(gene_universe)
        self.path = cached_path(url, cache_dir=cache_root / "pc")
        self._relations_by_pair = self._get_statements_by_pair()

    def get_statements(self, p1: str, p2: str) -> List[indra.statements.Statement]:
        return self._relations_by_pair.get((p1, p2), [])

    def _get_statements_by_pair(self) -> Dict[Tuple[str, str], List[statements.Statement]]:
        statements_by_pair = defaultdict(list)

        with gzip.open(self.path) as f:
            processor = process_owl_str(f.read().decode())
            for stmt in tqdm(processor.statements, desc=f"Indexing {self.name}"):
                if isinstance(stmt, statements.Modification):
                    if "EGID" in stmt.enz.db_refs and "EGID" in stmt.sub.db_refs:
                        enz = stmt.enz.db_refs["EGID"]
                        sub = stmt.sub.db_refs["EGID"]
                        if sub in self.gene_universe:
                            statements_by_pair[(enz, sub)].append(stmt)

        return statements_by_pair


if __name__ == '__main__':
    PathwayCommonsDB("pid", gene_universe={"5970", "8517"})




import abc
import os
import logging
import warnings
from collections import defaultdict
from typing import Optional, Set, List, Dict

import bioc
import diskcache
import numpy as np
import requests
from elasticsearch import Elasticsearch
from lxml import etree
from omegaconf import OmegaConf
from tqdm import tqdm

from pedl.utils import get_homologue_mapping, cache_root, SegtokSentenceSplitter, \
    cached_path, unpack_file, chunks, Entity, Sentence, get_pmid, \
    replace_consistently_dict, doc_synced


class DataGetter(abc.ABC):

    @abc.abstractmethod
    def get_sentences(self, head: Entity, tail: Entity) -> List[Sentence]:
        pass


class DataGetterPubtator(DataGetter):

    def __init__(self, elasticsearch, entity_marker: dict = None, entity_to_mask: dict = None,
                 max_size: int = 1000):
        # TODO check whether elastic search is running with pubtator index
        if elasticsearch.server.startswith("https://"):
            scheme, host, port = elasticsearch.server.split(":")
            host = host[2:]
        else:
            host, port = elasticsearch.server.split(":")
        self.types_to_blind = {"gene"}
        self.client = Elasticsearch(hosts=[{"host": host,
                                            "port": int(port),
                                            "scheme": "https",
                                            }], timeout=3000,
                                    basic_auth=("elastic", elasticsearch.password),
                                    ca_certs=elasticsearch.ca_certs)
        if entity_marker:
            self.et = entity_marker
        else:
            self.et = {"head_start": '<e1>',
                       "head_end": '</e1>',
                       "tail_start": '<e2>',
                       "tail_end": '</e2>'}

        self.entity_to_mask = entity_to_mask
        self.max_size = max_size

    def get_sentences(self, head: Entity, tail: Entity):
        processed_sentences = []
        query = {
            "bool": {
                "must": [
                    {"term": {head.type.lower(): head.cuid}},
                    {"term": {tail.type.lower(): tail.cuid}},
                ]
            }
        }
        result = self.client.search(query=query, index="pubtator_masked", size=self.max_size)
        retrieved_sentences = result["hits"]["hits"]
        for sentence in retrieved_sentences:
            sentence = sentence["_source"]

            spans_head = sentence["entities"][str(head)]
            spans_tail = sentence["entities"][str(tail)]
            spans_head_masked = sentence["entities_masked"][str(head)]
            spans_tail_masked = sentence["entities_masked"][str(tail)]

            text_orig = sentence["text"]
            text_orig_masked = sentence["text_masked"]

            for idx_head, span_head in enumerate(spans_head):
                for idx_tail, span_tail in enumerate(spans_tail):
                    span_to_replacement = {
                        tuple(span_head): self.et['head_start'] + text_orig[span_head[0]:span_head[-1]]
                        + self.et['head_end'],
                        tuple(span_tail): self.et['tail_start'] + text_orig[span_tail[0]:span_tail[-1]]
                        + self.et['tail_end'],
                    }
                    text, _ = replace_consistently_dict(text=text_orig, span_to_replacement=span_to_replacement)

                    span_head_masked = spans_head_masked[idx_head]
                    span_tail_masked = spans_tail_masked[idx_tail]

                    span_to_replacement = {
                        tuple(
                            span_head_masked): self.et['head_start'] + text_orig_masked[span_head_masked[0]:span_head_masked[1]] + self.et['head_end'],
                        tuple(
                            span_tail_masked): self.et['tail_start'] + text_orig_masked[span_tail_masked[0]:span_tail_masked[1]] + self.et['tail_end'],
                    }
                    text_masked, _ = replace_consistently_dict(text=text_orig_masked,
                                                            span_to_replacement=span_to_replacement)

                    processed_sentences.append(
                        Sentence(text=text,
                                 text_blinded=text_masked,
                                 pmid=sentence["pmid"],
                                 start_pos=0,
                                 entity_marker=self.et
                                 )
                    )

        return processed_sentences


class DataGetterAPI(DataGetter):
    CHUNK_SIZE = 100

    def __init__(
            self,
            gene_universe: Optional[Set[str]] = None,
            chemical_universe: Optional[Set[str]] = None,
            expand_species: Optional[List[str]] = None,
            entity_to_mask: Optional[Dict[str, str]] = None,
            entity_marker: dict = None
    ):
        self.gene_universe = gene_universe or set()
        self.chemical_universe = chemical_universe or set()
        self.expand_species = expand_species or []
        self.entity_to_mask = entity_to_mask or set()
        if self.expand_species:
            self.homologue_mapping = get_homologue_mapping(
                self.expand_species, self.gene_universe
            )
        else:
            self.homologue_mapping = {}
        self.gene2pmid = self.get_gene2pmid()
        self.chemical2pmid = self.get_chemical2pmid()
        self._document_cache = diskcache.Cache(
            directory=str(cache_root / "document_cache"),
            eviction_policy="least-recently-used",
        )
        if entity_marker:
            self.entity_marker = entity_marker
        else:
            self.entity_marker = {"head_start": '<e1>',
                                  "head_end": '</e1>',
                                  "tail_start": '<e2>',
                                  "tail_end": '</e2>'}
        self.sentence_splitter = SegtokSentenceSplitter(self.entity_marker)

    def get_gene2pmid(self):
        gene2pmid = defaultdict(set)

        final_path = cache_root / "data" / "gene2pubtatorcentral"
        if not final_path.exists():
            print("Downloading gene2pubtatorcentral...")
            path = cached_path(
                "https://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/gene2pubtatorcentral.gz",
                "data",
            )
            unpack_file(path, final_path)
        doc_synced(final_path, 'gene')

        with final_path.open() as f:
            for line in tqdm(f, total=70000000, desc="Loading gene2pubtatorcentral"):
                line = line.strip()
                fields = line.split("\t")
                gene_id = fields[2]
                pmid = fields[0]
                normalizers = fields[4]
                if "GNormPlus" not in normalizers:
                    continue

                if gene_id in self.gene_universe:
                    gene2pmid[gene_id].add(pmid)
                elif gene_id in self.homologue_mapping:
                    for mapped_gene_id in self.homologue_mapping[
                        gene_id
                    ]:  # mapped_gene_id is from self.protein_universe
                        gene2pmid[mapped_gene_id].add(pmid)

        return dict(gene2pmid)

    def get_chemical2pmid(self):
        chemical2pmid = defaultdict(set)

        if not self.chemical_universe:
            return chemical2pmid

        final_path = cache_root / "data" / "chemical2pubtatorcentral"
        if not final_path.exists():
            print("Downloading chemical2pubtatorcentral...")
            path = cached_path(
                "https://ftp.ncbi.nlm.nih.gov/pub/lu/PubTatorCentral/chemical2pubtatorcentral.gz",
                "data",
            )
            unpack_file(path, final_path)

        doc_synced(final_path, 'chemical')
        with final_path.open() as f:
            for line in tqdm(
                    f, total=104567794, desc="Loading chemical2pubtatorcentral"
            ):
                line = line.strip()
                fields = line.split("\t")
                chemical_id = fields[2]
                pmid = fields[0]
                normalizers = fields[4]
                if "TaggerOne" not in normalizers:
                    continue

                if chemical_id in self.chemical_universe:
                    chemical2pmid[chemical_id].add(pmid)
        return dict(chemical2pmid)

    def maybe_map_to_pmcid(self, pmids):
        pmid_to_pmcid = {}

        service_root = "https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/"
        for pmid_chunk in chunks(pmids, 200):
            ids = ",".join(pmid_chunk)
            response = requests.get(service_root, params={"ids": ids})
            try:
                for record in etree.fromstring(response.content).xpath("//record"):
                    if "pmcid" in record.attrib:
                        pmid_to_pmcid[record.attrib["pmid"]] = record.attrib["pmcid"]
            except etree.XMLSyntaxError:
                warnings.warn(
                    "Failed to parse PubTator response."
                    " You are probably issuing too many requests."
                    " Please use a local copy of PubTator or disable api_fallback, if you already do."
                )

        return pmid_to_pmcid

    @staticmethod
    def get_entities_from_annotation(annotation: bioc.BioCAnnotation, homologue_mapping: Dict) -> Set[Entity]:
        """
        Extract entities from `annotation` and expand them with homologues from
        `self.expand_species` if it is a Gene.
        """
        if "identifier" in annotation.infons:
            identifiers = annotation.infons["identifier"]
            if not identifiers:
                identifiers = ""
        elif "Identifier" in annotation.infons:
            identifiers = annotation.infons["Identifier"]
            if not identifiers:
                identifiers = ""
        else:
            identifiers = ""

        identifiers = set(identifiers.split(";"))

        expanded_identifiers = identifiers.copy()
        if annotation.infons["type"] == "Gene":
            for cuid in identifiers:
                expanded_identifiers.update(homologue_mapping.get(cuid, {}))

        return set(Entity(cuid=cuid, type=annotation.infons["type"]) for cuid in expanded_identifiers)

    def get_sentences_from_document(self, entity1: Entity, entity2: Entity, document):
        sentences = []
        for passage in document.passages:
            entity1_locations = []
            entity1_lengths = []
            entity2_locations = []
            entity2_lengths = []
            for annotation in passage.annotations:
                if annotation.infons["type"] not in {entity1.type, entity2.type}:
                    continue

                entities_ann = self.get_entities_from_annotation(annotation, self.homologue_mapping)

                if entity1 in entities_ann:
                    for loc in annotation.locations:
                        entity1_locations.append(loc.offset - passage.offset)
                        entity1_lengths.append(loc.length)
                if entity2 in entities_ann:
                    for loc in annotation.locations:
                        entity2_locations.append(loc.offset - passage.offset)
                        entity2_lengths.append(loc.length)

            entity1_locations_arr = np.array(entity1_locations).reshape(-1, 1)
            entity2_locations_arr = np.array(entity2_locations).reshape(1, -1)
            dists = abs(entity1_locations_arr - entity2_locations_arr)
            for i, j in zip(*np.where(dists <= 300)):
                loc_ent1 = entity1_locations_arr[i, 0]
                loc_ent2 = entity2_locations_arr[0, j]
                len_ent1 = entity1_lengths[i]
                len_ent2 = entity2_lengths[j]
                sentence = self.get_sentence(
                    passage=passage,
                    offset_ent1=loc_ent1,
                    offset_ent2=loc_ent2,
                    len_ent1=len_ent1,
                    len_ent2=len_ent2,
                    pmid=get_pmid(document)[0],
                )
                if sentence:
                    sentences.append(sentence)

        return sentences

    def get_pmids(self, entity: Entity) -> Set[str]:
        if entity.type == "Chemical":
            cuid_to_pmid = self.chemical2pmid
        elif entity.type == "Gene":
            cuid_to_pmid = self.gene2pmid
        else:
            raise ValueError

        return cuid_to_pmid.get(entity.cuid, set())

    def get_sentences(self, head: Entity, tail: Entity) -> List[Sentence]:
        sentences = []
        pmids = sorted(self.get_pmids(head) & self.get_pmids(tail))
        if not pmids:
            return []

        for documents in self.get_documents_from_api(pmids):
            for document in documents:
                sentences += self.get_sentences_from_document(
                    entity1=head, entity2=tail, document=document
                )

        return sentences

    def cache_documents(self, documents: List[bioc.BioCDocument]) -> None:
        logging.info(f"Caching {len(documents)} documents")
        for document in documents:
            pmid = get_pmid(document)[0]
            self._document_cache[pmid] = document

    def get_documents_from_api(self, pmids):
        service_root = "https://www.ncbi.nlm.nih.gov/research/pubtator-api/publications/export/biocxml"
        pmids = list(pmids)

        if len(pmids) > self.CHUNK_SIZE:
            pbar = tqdm(desc="Collecting", total=len(pmids))
        else:
            pbar = None

        cached_pmids = [i for i in pmids if i in self._document_cache]

        for pmid_chunk in chunks(cached_pmids, self.CHUNK_SIZE):
            if pbar:
                pbar.update(len(pmid_chunk))
            yield [self._document_cache[i] for i in pmid_chunk]

        uncached_pmids = [i for i in pmids if i not in cached_pmids]
        pmid_to_pmcid = self.maybe_map_to_pmcid(uncached_pmids)

        pmids_to_retreive = [i for i in uncached_pmids if i not in pmid_to_pmcid]
        pmcids_to_retreive = [
            pmid_to_pmcid[i] for i in uncached_pmids if i in pmid_to_pmcid
        ]

        pmcid_to_pmid = {v: k for k, v in pmid_to_pmcid.items()}

        for pmid_chunk in list(chunks(pmids_to_retreive, self.CHUNK_SIZE)):

            result = requests.get(
                service_root, params={"pmids": ",".join(pmid_chunk), "concepts": "gene,chemical"}
            )
            collection = bioc.loads(result.content.decode())
            yield collection.documents
            if pbar:
                pbar.update(len(pmid_chunk))
            self.cache_documents(collection.documents)

        for pmcid_chunk in list(chunks(pmcids_to_retreive, self.CHUNK_SIZE)):

            result = requests.get(
                service_root,
                params={"pmcids": ",".join(pmcid_chunk), "concepts": "gene"},
            )
            collection = bioc.loads(result.content.decode())
            yield collection.documents
            if pbar:
                pbar.update(len(pmcid_chunk))
            self.cache_documents(collection.documents)

    def get_sentence(
            self,
            passage,
            offset_ent1,
            offset_ent2,
            len_ent1,
            len_ent2,
            pmid,
            allow_multi_sentence=False,
    ):

        if offset_ent1 < offset_ent2:
            left_start = offset_ent1
            right_start = offset_ent2
        else:
            left_start = offset_ent2
            right_start = offset_ent1
        sents = self.sentence_splitter.split(passage.text.strip())
        snippet_start = None
        snippet_end = None
        for sent in sents:
            if sent.end_pos >= left_start >= sent.start_pos:
                snippet_start = sent.start_pos

            if (
                    sent.end_pos >= right_start >= sent.start_pos
            ):  # is sentence after right entity
                snippet_end = sent.end_pos

        if snippet_start is not None and snippet_end is None:
            snippet_end = sent.end_pos
        if snippet_start is None:
            return None

        if not allow_multi_sentence and not any(
                snippet_start == i.start_pos and snippet_end == i.end_pos for i in sents
        ):
            return None  # is multi sentence

        text = passage.text[snippet_start:snippet_end]
        span_to_replacement = {}
        span_to_replacement_masked = {}
        masked_count = 0
        entity_to_spans = defaultdict(list)
        for ann in passage.annotations:
            for loc in ann.locations:
                if snippet_end >= loc.offset - passage.offset >= snippet_start:
                    entities = self.get_entities_from_annotation(ann, self.homologue_mapping)

                    start = loc.offset - passage.offset - snippet_start
                    end = loc.end - passage.offset - snippet_start
                    for entity in entities:
                        entity_to_spans[entity].append((start, end))

        for entity, spans in entity_to_spans.items():
            masked_count += 1
            for start, end in spans:
                ent_text = text[start:end]
                if entity.type in self.entity_to_mask:
                    replacement_masked = f"<{self.entity_to_mask[entity.type]}{masked_count}/>"
                else:
                    replacement_masked = ent_text
                replacement = ent_text

                if start == offset_ent1 - snippet_start:
                    replacement = f"{self.entity_marker['head_start']}{replacement}{self.entity_marker['head_end']}"
                    replacement_masked = f"{self.entity_marker['head_start']}{replacement_masked}{self.entity_marker['head_end']}"

                if start == offset_ent2 - snippet_start:
                    replacement = f"{self.entity_marker['tail_start']}{replacement}{self.entity_marker['tail_end']}"
                    replacement_masked = f"{self.entity_marker['tail_start']}{replacement_masked}{self.entity_marker['tail_end']}"
                span_to_replacement[(start, end)] = replacement
                span_to_replacement_masked[(start, end)] = replacement_masked

        text_marked, _ = replace_consistently_dict(text, span_to_replacement)
        text_masked, _ = replace_consistently_dict(text, span_to_replacement_masked)

        if self.entity_marker['head_start'] not in text_marked or self.entity_marker['head_end'] not in text_marked:
            return None # Weird encoding error

        return Sentence(pmid=pmid, text=text_marked, text_blinded=text_masked, start_pos=snippet_start, entity_marker=self.entity_marker)



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name='pedl',
    version='1.0.1',
    description='Search the biomedical literature for protein interactions and'
                'protein associations.',
    url='https://github.com/leonweber/pedl',
    author='Leon Weber',
    author_email='leonweber@posteo.de',
    license='MIT',
    packages=find_packages(),
    install_requires=required,
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["pedl-summarize=pedl.summarize:summarize",
                                      "pedl-rebuild_pubtator_index=pedl.rebuild_pubtator_index:rebuild_pubtator_index",
                                      "pedl-build_training_set=pedl.build_training_set:build_training_set",
                                      "pedl-extract=pedl.predict:predict"]},
    package_data={"pedl": ["data/*",
                           'configs/*',
                           'configs/database/*',
                           'configs/entities/*',
                           'configs/elastic/*',
                           'configs/type/*',
                           'configs/hydra/help/*']}
)



import string
from collections import defaultdict
from operator import itemgetter
from typing import Optional, Dict, List, Set, Tuple
from argparse import ArgumentParser
from pathlib import Path
import random

from lxml import etree
from openpyxl.styles import Font, PatternFill
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl import Workbook
import pandas as pd
import numpy as np
import requests
from openpyxl.worksheet.table import Table, TableStyleInfo
from omegaconf import DictConfig
from tqdm import tqdm, trange
import hydra
from Bio import Entrez

np.random.seed(42)
random.seed(42)

ESEARCH_MAX_COUNT = 100


def get_pmid_to_mesh_terms(pmids: Set[str], mesh_terms: Set[str]) -> Dict[str, List[str]]:
    """
    Get a mapping of PubMed IDs to their associated MeSH terms.

    This function retrieves MeSH terms for a given set of PubMed IDs (pmids) by querying the PubMed database.
    It then filters the MeSH terms based on a provided set of terms (mesh_terms) and returns a dictionary
    mapping each PubMed ID to a list of its associated MeSH terms.

    Args:
    pmids (Set[str]): A set of unique PubMed IDs for which MeSH terms are to be fetched.
    mesh_terms (Set[str]): A set of MeSH terms to filter the results by.

    Returns:
    Dict[str, List[str]]: A dictionary mapping each PubMed ID (key) to a list of its associated MeSH terms (value).

    Examples:
    >>> pmids = {"12345678", "23456789"}
    >>> mesh_terms = {"Disease", "Therapeutics"}
    >>> get_pmid_to_mesh_terms(pmids, mesh_terms)
    {'12345678': ['Disease'], '23456789': ['Therapeutics']}
    """
    pmid_to_mesh_terms = defaultdict(list)

    pmids_list = list(pmids)
    for i in trange(0, len(pmids_list), ESEARCH_MAX_COUNT, desc="Fetching MeSH terms"):
        batch_pmids = pmids_list[i:i + ESEARCH_MAX_COUNT]
        pmids_str = ",".join(batch_pmids)

        Entrez.email = "leonweber@cis.lmu.de"
        Entrez.tool = "PEDL"
        handle = Entrez.efetch(db="pubmed", id=pmids_str, retmode="xml")
        tree = etree.parse(handle)

        for article in tree.xpath("//PubmedArticle"):
            pmid = article.xpath("MedlineCitation/PMID")[0].text
            mesh_terms_pmid = set(article.xpath("MedlineCitation/MeshHeadingList/MeshHeading/DescriptorName/text()"))
            mesh_terms_pmid = mesh_terms_pmid.intersection(mesh_terms)  # TODO We might change this to also match subterms

            pmid_to_mesh_terms[pmid] = sorted(mesh_terms_pmid)

    return pmid_to_mesh_terms


def adjust_sheet_width(sheet):
    # Adjust column width
    for col in sheet.columns:
        max_length = 0
        column = col[0].column_letter  # Get the column name
        for cell in col:
            try:  # Necessary to avoid error on empty cells
                if len(str(cell.value)) > max_length:
                    max_length = len(str(cell.value))
            except:
                pass
        adjusted_width = min((max_length + 2) * 1.2, 255)
        sheet.column_dimensions[column].width = adjusted_width


def _add_value_choices(sheet):
    # Incorrect
    sheet["P1"].value = "Wrong Genes"
    sheet["P2"].value = "Wrong PPA Type"
    sheet["P3"].value = "No PPA"

    data_validation = DataValidation(type="list", formula1="$P$1:$P$100")
    data_validation.add("K2:K10000")
    sheet.add_data_validation(data_validation)

    # Useless
    sheet["O1"].value = "Indirect Interaction"
    sheet["O2"].value = "Insufficient Evidence"
    sheet["O3"].value = "Wrong Organism"
    sheet["O4"].value = "Wrong Disease"
    sheet["O5"].value = "Mutation Required"
    sheet["O6"].value = "PTM Required"
    sheet["O7"].value = "Bound Proteins Required"

    data_validation = DataValidation(type="list", formula1="$O$1:$O$100")
    data_validation.add("L2:L10000")
    sheet.add_data_validation(data_validation)


def _add_table(sheet, num_rows, num_cols):
    last_col = string.ascii_uppercase[num_cols - 1]
    ref = f"A1:{last_col}{num_rows}"
    i = 1
    while True:
        try:
            table = Table(displayName=f"Table{i}", ref=ref)
            table.tableStyleInfo = TableStyleInfo(
                name="TableStyleLight1",
                showFirstColumn=False,
                showLastColumn=False,
                showRowStripes=True,
                showColumnStripes=True,
            )
            sheet.add_table(table)
            break
        except ValueError:
            i += 1


def fill_sheet_from_df(
    sheet,
    df: pd.DataFrame,
    ppa_dir: Path,
    top_k_articles: int = 5,
    threshold: float = 0.5,
    pmid_to_mesh_terms: Optional[Dict] = None,
    annotation_mode: bool = False,
):
    header = [
        "head",
        "association type",
        "tail",
        "text",
        "pubmed",
        "article score",
        "total score",
        "mean score",
        "MESH terms",
    ]
    if annotation_mode:
        header += ["correct", "useful", "why incorrect?", "why not useful?", "comment"]
    for cell, value in zip(sheet["A1":"Z1"][0], header):
        cell.value = value
        cell.font = Font(bold=True)

    idx_next_free_row = 0

    all_ppa_data = []

    for i, (_, row) in enumerate(df.iterrows()):
        ppa_data = {"row": row, "score": 0}
        fname = ppa_dir / ((row["head"] + "-_-" + row["tail"]).upper() + ".txt")
        assert fname.exists(), f"{fname} does not exist"
        ppa_data["pmid_to_score"] = defaultdict(float)
        ppa_data["pmid_to_fields"] = defaultdict(list)

        with fname.open() as f:
            for line in f:
                fields = line.strip().split("\t")
                if len(fields) > 1:
                    ppa_type, score, pmid, text, _ = fields
                    if float(score) < threshold or ppa_type != row["association type"]:
                        continue

                    if pmid_to_mesh_terms and not pmid_to_mesh_terms[pmid]:
                        continue

                    ppa_data["pmid_to_score"][pmid] += float(score)
                    ppa_data["pmid_to_fields"][pmid].append(fields)
                    ppa_data["score"] += float(score)
        all_ppa_data.append(ppa_data)

    for ppa_data in sorted(all_ppa_data, key=itemgetter("score"), reverse=True):
        for i, (pmid, article_score) in enumerate(sorted(
            ppa_data["pmid_to_score"].items(), key=itemgetter(1), reverse=True
        )[:top_k_articles]):
            fields = ppa_data["pmid_to_fields"][pmid][0]
            ppa_type, score, pmid, text, _ = fields
            sheet[f"A{idx_next_free_row + 2}"].value = ppa_data["row"]["head"]
            sheet[f"C{idx_next_free_row + 2}"].value = ppa_data["row"]["tail"]
            sheet[f"B{idx_next_free_row + 2}"].value = ppa_data["row"][
                "association type"
            ]
            sheet[f"D{idx_next_free_row + 2}"].value = text
            sheet[
                f"E{idx_next_free_row + 2}"
            ].hyperlink = f"https://pubmed.ncbi.nlm.nih.gov/{pmid}/"
            sheet[f"E{idx_next_free_row + 2}"].value = pmid
            sheet[f"B{idx_next_free_row + 2}"].value = ppa_data["row"][
                "association type"
            ]
            sheet[f"E{idx_next_free_row + 2}"].style = "Hyperlink"
            sheet[f"F{idx_next_free_row + 2}"].value = article_score
            if i == 0:
                sheet[f"F{idx_next_free_row + 2}"].font = Font(bold=True)

            sheet[f"G{idx_next_free_row + 2}"].value = ppa_data["row"]["score (sum)"]

            sheet[f"H{idx_next_free_row + 2}"].value = ppa_data["row"]["score (sum)"] / ppa_data["row"]["num_found"]

            if pmid_to_mesh_terms and pmid in pmid_to_mesh_terms:
                sheet[f"I{idx_next_free_row + 2}"].value = ", ".join(
                    pmid_to_mesh_terms[pmid]
                )
            idx_next_free_row += 1

    if annotation_mode:
        _add_value_choices(sheet)
    adjust_sheet_width(sheet)
    if idx_next_free_row > 0:
        _add_table(sheet, num_rows=idx_next_free_row + 1, num_cols=len(header))


def _get_ppas_from_sheet(sheet):
    ppas = set()

    it = zip(sheet["A"], sheet["B"], sheet["C"])
    next(it) # skip header
    for head, rel, tail in it:
        ppa = (head.value, rel.value, tail.value)
        if all(ppa):
            ppas.add(ppa)

    return ppas



def _add_summary_sheet(df_summary, sheet, wb):
    ppas_in_sheet = _get_ppas_from_sheet(sheet)
    summary_sheet = wb.create_sheet(title=("Summary_" + sheet.title)[:31])
    n_rows_added = 0
    for i, row in enumerate(
            dataframe_to_rows(df_summary.sort_values("score (sum)", ascending=False),
                              index=False, header=True)):
        if i == 0 or tuple(row[:3]) in ppas_in_sheet:
            summary_sheet.append(row)
            n_rows_added += 1
    adjust_sheet_width(summary_sheet)
    _add_table(summary_sheet, num_rows=n_rows_added,
               num_cols=len(df_summary.columns))


def build_summary_table(
    raw_dir: Path, threshold: float = 0.0, no_association_type: bool = False
) -> pd.DataFrame:
    df = {
        "head": [],
        "association type": [],
        "tail": [],
        "score (sum)": [],
        "score (max)": [],
        "pmids": [],
        "num_found": [],
    }


    rel_to_score_sum = defaultdict(float)
    rel_to_score_max = defaultdict(float)
    rel_to_num_found = defaultdict(int)
    rel_to_pmids = defaultdict(set)

    # hgnc_symbols = set(get_hgnc_symbol_to_gene_id().keys())

    files = list(raw_dir.glob("*.txt"))
    for file in tqdm(files):
        with file.open() as f:
            p1, p2 = file.name.replace(".txt", "").split("-_-")
            for line in f:
                fields = line.strip().split()
                if fields:
                    if no_association_type:
                        p1_unified, p2_unified = sorted([p1, p2])
                        rel = (p1_unified, "association", p2_unified)
                    else:
                        rel = (p1, fields[0], p2)
                    if float(fields[1]) >= threshold:
                        rel_to_score_sum[rel] += float(fields[1])
                        rel_to_num_found[rel] += 1
                        rel_to_score_max[rel] = max(
                            float(fields[1]), rel_to_score_max[rel]
                        )
                        rel_to_pmids[rel].add(fields[2])

    for rel, score_sum in rel_to_score_sum.items():
        score_max = rel_to_score_max[rel]
        pmids = ",".join(rel_to_pmids[rel])
        df["head"].append(rel[0])
        df["association type"].append(rel[1])
        df["tail"].append(rel[2])
        df["score (sum)"].append(score_sum)
        df["score (max)"].append(score_max)
        df["pmids"].append(pmids)
        df["num_found"].append(rel_to_num_found[rel])

    return pd.DataFrame(df)


def summarize_excel(cfg):
    df_summary = build_summary_table(raw_dir=Path(cfg.input), threshold=cfg.threshold)
    all_pmids = set()
    for pmids_rel in df_summary["pmids"]:
        all_pmids.update(pmids_rel.split(","))
    if cfg.mesh_terms:
        pmid_to_mesh_terms = get_pmid_to_mesh_terms(all_pmids, set(cfg.mesh_terms))
    else:
        pmid_to_mesh_terms = None

    wb = Workbook()
    sheet = wb.active

    if cfg.entity_set:
        with cfg.entity_set.open() as f:
            entity_set = set(f.read().split("\n"))
            df_a_to_b = df_summary[df_summary["head"].isin(entity_set)]
            df_b_to_a = df_summary[df_summary["tail"].isin(entity_set)]
            sheet.title = f"{cfg.entity_set.with_suffix('').name} -> other"[:31]
            fill_sheet_from_df(
                sheet=sheet,
                df=df_a_to_b,
                ppa_dir=Path(cfg.input),
                threshold=cfg.threshold,
                top_k_articles=cfg.top_k_articles,
                pmid_to_mesh_terms=pmid_to_mesh_terms,
            )
            _add_summary_sheet(df_a_to_b, sheet, wb)

            sheet = wb.create_sheet(title=f"other -> {cfg.entity_set.with_suffix('').name}"[:31])
            fill_sheet_from_df(
                sheet=sheet,
                df=df_b_to_a,
                ppa_dir=Path(cfg.input),
                threshold=cfg.threshold,
                top_k_articles=cfg.top_k_articles,
                pmid_to_mesh_terms=pmid_to_mesh_terms,
            )
            _add_summary_sheet(df_b_to_a, sheet, wb)
    else:
        fill_sheet_from_df(
            sheet=sheet,
            df=df_summary,
            ppa_dir=Path(cfg.input),
            threshold=cfg.threshold,
            top_k_articles=cfg.top_k_articles,
            pmid_to_mesh_terms=pmid_to_mesh_terms,
        )

        _add_summary_sheet(df_summary, sheet, wb)

    output_file = Path(cfg.output).with_suffix(".xlsx")
    output_file.parent.mkdir(parents=True, exist_ok=True)
    wb.save(output_file)


@hydra.main(config_path="configs", config_name="summarize.yaml", version_base=None)
def summarize(cfg: DictConfig):
    summarize_excel(cfg)

if __name__ == "__main__":
    summarize()